\ YDOWN - YMODEM sender
\ Pavel Zampach, 2021 (pavelzampach@gmail.com)
\ Source text for DX-Forth
\ Version 1.5
\ Use YDOWN.CFG for set LED blinking, 1k blocks and CPU speed
\ *

decimal

$01 constant SOH
$02 constant STX
$04 constant EOT
$06 constant ACK
$15 constant NAK
$18 constant CAN

\ Buffer structure:
\   0   |  1   |  2 ... 129/1025      | 130/1026
\ ------+------+----------------------+---------
\ blockS|blockN| data 128/1024 bytes  | csum
\ ^      ^      ^
\ blks   blkn   buf

$8000 set-limit
$8000   constant BLKS
blks 1+ constant BLKN
blkn 1+ constant BUF
$8480   constant CMD                    \ copy of command line tail

0 value DATLEN                          \ data length (128 or 1024)
0 value HBYTE                           \ leading byte of block (SOH or STX)
variable WNGCOUNT                       \ wrong block counter
create #REC   3 allot                   \ record counter (third byte is not used)
false value OPENERR                     \ open file error
0 value FSEND                           \ file id of sending file
0 value FLIST                           \ file id of file list

$010b constant 1KBLOCK                  \ addr of params
$010c constant LEDPORT
$010d constant LED+
$010d constant LED-


\ Jump to CONOUT
label OUTBIOS
  0 jmp                                 \ will be replaced by CONOUT addr
end-code

\ Wait 1 s
: WAIT        ( -- )
  1000 ms
;

\ Switch LED on
: LEDON       ( -- )
  ledport c@ ?dup if
    led+ c@ swap pc!
  then
;

\ Switch LED off
: LEDOFF      ( -- )
  ledport c@ ?dup if
    led- c@ swap pc!
  then
;

\ Print msg [addr u] and end program
: ERR         ( adr u -- )
  CAN emit   wait                       \ cancel transfer
  type cr quit
;

\ Test I/O result
: FILE-CHECK  ( ior -- )
  if s" File access error" err then
;

\ Initialisation
: INIT        ( -- )
  cmdtail cmd place                     \ store command line
  cmd count upper                       \ and convert to uppercase
  1 @ 9 + outbios 1+ !                  \ init BIOS entry point (CONOUT)
;

\ Checksum [datlen] bytes and store to CSUM
code CSUM     ( datlen -- )
   buf H lxi
       D pop    \ datlen to DE
       B push   \ save BC (forth interpretive pointer)
       A xra    \ null A
1 $:   M add
       H inx
       D dcx
     A C mov    \ transfer checksum
     D A mov
       E ora
     C A mov    \ restore checksum
     1 $ jnz
     A M mov    \ checksum to CSUM
       B pop
         next   \ return to forth
end-code

\ Write [blklen] bytes from [blks] to console
code WRITEBLK ( blklen -- )
  blks H lxi
       D pop    \ blklen to DE
       B push   \ save BC (forth interpretive pointer)
1 $: M C mov
       H push
       D push
 outbios call   \ call CONOUT
       D pop
       H pop
       H inx
       D dcx
     D A mov
       E ora
     1 $ jnz
       B pop
         next   \ return to forth
end-code

\ Set params for 128-byte block
: BLK1REC    ( -- )
  128 to datlen
  SOH to hbyte
;

\ Set params for 1024-byte block
: BLK8REC     ( -- )
  1024 to datlen
  STX  to hbyte
;

\ Prepare zero block
: ZEROBLK     ( -- )
  blk1rec
  buf datlen erase
  $ff00 blks !                          \ 0 -> blockS, $ff -> blockN
;

\ Set block params
: SETBLKPAR   ( -- )
  #rec @ 7 u> 1kblock c@ and if         \ test record counter and 1Kblock flag
    blk8rec -8
  else
    blk1rec -1
  then
  #rec +!                               \ decrement record counter
  ;c
    blks H lxi   M inr   H inx   M dcr  \ blockS++, blockN--
  c:
;

\ Send block
: SENDBLK     ( -- )
  ledon
  datlen csum                           \ make and store csum
  hbyte emit   datlen 3 + writeblk      \ write block
  ledoff
;

\ Read and send next data block
: DATABLK     ( -- )
  setblkpar
  buf datlen fsend read-file file-check drop
  sendblk
;

\ Test wrong block counter
: TESTWRONG   ( -- )
  wngcount @ 3 > if
    s" Too many errors" err
  then
;

\ Send file named [addr u], path optional
: SEND-FILE   ( addr u -- )
  wngcount off   zeroblk                \ init
  2dup
  r/o open-file throw to fsend
  -path buf swap cmove                  \ filename to buffer
  fsend file-size file-check #rec rec!  \ set record counter
  sendblk                               \ send zero block with filename

  begin
    #rec @
  while
    key case
      ACK of wngcount off    datablk endof
      NAK of 1 wngcount +!   sendblk endof
      CAN of s" Canceled" err endof
      drop
    endcase
    testwrong
  repeat

  fsend close-file file-check
  EOT emit   key drop                   \ NAK
  EOT emit   key drop                   \ ACK
;

\ Return the first substring [addr3 u3] from string [addr1 u1]
\ separated by char [c], string [addr2 u2] is the remainder. When
\ [addr3 u3] is the last part of original string, u2 = -1
: SLICE       ( addr1 u1 c  -- addr2 u2 addr3 u3 )
  >r 2dup r> scan dup >r
  1 /string 2swap r> -
;

\ Send files from file list [addr u]
: LISTSEND    ( addr u -- )
  r/o open-file throw to flist
  begin
    pad 20 flist read-line file-check
  while
    pad swap 2dup upper                 \ convert filename to upper case
    ['] send-file catch if              \ if file can't be opened, skip it over
      2drop   true to openerr
    then
  repeat   drop
  flist close-file file-check
;

\ Send files from command line
: CLINESEND   (  -- )
  cmd count
  begin
    bl skip bl slice                    \ skip leading spaces, fetch filename
    2dup s" .YFL" search nip nip
    if ['] listsend else ['] send-file then
    catch if                            \ if file can't be opened, skip it over
      2drop   true to openerr
    then
    dup 0<                              \ if was the last filename, end
  until   2drop
  zeroblk sendblk                       \ end of session
;

: MAIN        ( -- )
  cmdtail nip if
    init
    ." YDOWN 1.5 - start recieve" cr
    key NAK <> abort" Interrupted"      \ reciever starts by NAK
    clinesend
    wait ." Download completed" cr
    openerr if ." Some files wasn't found" cr then
  else
    cr ." Using YDOWN ufn|fn.YFL{ ufn|fn.YFL}"
  then
;

turnkey main ydown
