\ YUP - YMODEM reciever
\ Pavel Zampach, 2021 (pavelzampach@gmail.com)
\ Source text for DX-Forth
\ Version 1.5
\ Use YUP.CFG for set LED blinking and CPU speed
\ *

decimal

$01 constant SOH
$02 constant STX
$06 constant ACK
$15 constant NAK
$18 constant CAN

\ Buffer structure:
\   0   |  1   |  2 ... 129/1025      | 130/1026
\ ------+------+----------------------+---------
\ blockS|blockN| data 128/1024 bytes  | csum
\ ^      ^      ^                      ^
\ blks   blkn   buf                    sum

$8000 set-limit
$8000   constant BLKS
blks 1+ constant BLKN
blkn 1+ constant BUF
$8480   constant CMD                    \ copy of command line tail
128     constant RECLEN                 \ record length

0 value DATLEN                          \ data length (128 or 1024)
0 value SUM                             \ checksum
variable BLKCOUNT                       \ block counter
variable WNGCOUNT                       \ wrong block counter
variable ANSWER                         \ ACK/NAK
0 value FID                             \ file id

$010c constant LEDPORT                  \ addr of params
$010d constant LED+
$010d constant LED-


\ Jump to CONIN
label INBIOS
  0 jmp                                 \ will be replaced by CONIN addr
end-code

\ Wait 1 s
: WAIT        ( -- )
  1000 ms
;

\ Switch LED on
: LEDON       ( -- )
  ledport c@ ?dup if
    led+ c@ swap pc!
  then
;

\ Switch LED off
: LEDOFF      ( -- )
  ledport c@ ?dup if
    led- c@ swap pc!
  then
;

\ Print msg [addr u] and abort program
: ERR         ( addr u -- )
  CAN emit   wait                       \ cancel transfer
  type cr abort
;

\ Test I/O result
: FILE-CHECK  ( ior -- )
  if s" File access error" err then
;

\ Initialisation
: INIT        ( -- )
  cmdtail cmd place                     \ store command line
  1 @ 6 + inbios 1+ !                   \ init BIOS entry point
;

\ Convert null terminated string
: ZCOUNT      ( addr -- addr u )
  dup -1 0 scan drop over -
;

\ Build filename from path on command line and name in buffer
: FNAME       ( -- addr u )
  buf zcount
  cmd count >pad +string
;

\ Checksum [datlen] bytes from [buf]
code CSUM     ( datlen -- csum )
   buf H lxi
       D pop    \ datlen to DE
       B push   \ save BC (forth interpretive pointer)
       A xra    \ null A
1 $:   M add
       H inx
       D dcx
     A C mov    \ transfer csum
     D A mov
       E ora
     C A mov    \ restore csum
     1 $ jnz
     0 H mvi    \ csum to HL
     A L mov
       B pop
         1push  \ return to forth pushing HL onto stack
end-code

\ Read [blklen] bytes from console to [blks]
code READBLK  ( blklen -- )
  blks H lxi
       D pop    \ blklen to DE
       B push   \ save BC (forth interpretive pointer)
1 $:   H push
       D push
  inbios call   \ call CONIN
       D pop
       H pop
     A M mov
       H inx
       D dcx
     D A mov
       E ora
     1 $ jnz
       B pop
         next   \ return to forth
end-code

\ Set params for block, [u] = data length (128/1024)
: BLKPARA     ( u -- )
  to datlen
  buf datlen + to sum
;

\ Check block, flag = block OK
: BLKCHK?     ( -- flag )
  blks c@ blkcount c@ =                 \ check blockS
  blks c@ blkn c@ + $ff = and           \ check blockS + blockN
  datlen csum sum c@ = and              \ check sum
;

\ Recieve block w/o header byte and check it, set answer, flag = block OK
: RCVBLK?     ( -- flag )
  ledon
  datlen 3 + readblk
  blkchk? dup if                        \ block OK
    1 blkcount +!
    wngcount off
    ACK
  else                                  \ wrong block
    1 wngcount +!
    NAK
  then
  answer !
  ledoff
;

\ Test wrong block counter and send no-null answer
: SENDANS     ( -- )
  wngcount @ 3 > if
    s" Too many errors" err
  then
  answer @ ?dup if emit then
;

\ Init variables and read zero block
: ZEROBLK     ( -- )
  blkcount off   answer off   wngcount off
  reclen blkpara                        \ zero block is always 128-byte
  begin
    sendans
    key SOH <> if                       \ read and test header byte
      s" Interrupted" err
    then
    rcvblk?
  until
;

\ Set block params, flag -> c = SOH or c = STX
: HEAD?       ( c -- flag)
  dup
  SOH = dup if reclen blkpara then
  swap
  STX = dup if [ reclen 8 * ] literal blkpara then
  or
;

\ Read blocks to end of file
: READTOEND   ( -- )
  begin
    key head?
  while
    rcvblk? if
      buf datlen fid write-file file-check
    then
    sendans
  repeat
;

: SESSION     ( -- )
  begin
    zeroblk                             \ init and read zero block
    buf c@                              \ empty zero block -> end of session
  while
    fname w/o create-file file-check to fid
    sendans                             \ answer for previous block
    readtoend
    fid close-file file-check
    NAK emit   key drop   ACK emit      \ end of file transfer
  repeat
  sendans                               \ answer for terminate zero block
;

: MAIN        ( -- )
  ." YUP 1.5 - start upload" cr
  init
  session
  wait ." Upload completed" cr
;

turnkey main yup
